/*
 * $Id$
 * Copyright (C) TOMITA Masahiro <mailto:tommy@tmtm.org>
 */
#include <stdarg.h>
#include <errno.h>
#include <ruby.h>
#include <senna/senna.h>

struct index_data {
    int closed;
    sen_index *indexp;
};

static void free_index(struct index_data *data)
{
    if (!data->closed) {
        sen_index_close(data->indexp);
        xfree(data);
    }
}

static VALUE mSenna;
static VALUE cSennaIndex;
static VALUE cSennaRecords;
static VALUE cSennaValues;
static VALUE eSennaError;
static VALUE eSennaMemoryExhausted;
static VALUE eSennaInvalidFormat;
static VALUE eSennaFileOperationError;
static VALUE eSennaInvalidArgument;
static VALUE eSennaOtherError;
static VALUE eExternalError;
static VALUE eInternalError;
static VALUE eAbnormalError;
static VALUE eEndOfData;

static VALUE *exceptions[] = {
    &eSennaError,
    &eSennaMemoryExhausted,
    &eSennaInvalidFormat,
    &eSennaFileOperationError,
    &eSennaInvalidArgument,
    &eSennaOtherError,
    &eExternalError,
    &eInternalError,
    &eAbnormalError,
    &eEndOfData,
};

static void senna_error(sen_rc rc, const char *fmt, ...)
{
    va_list args;
    char buf[BUFSIZ];
    VALUE ec, e;

    va_start(args, fmt);
    vsnprintf(buf, BUFSIZ, fmt, args);
    va_end(args);
    if (rc < sizeof(exceptions)/sizeof(exceptions[0]))
        ec = *exceptions[rc];
    else
        ec = eSennaError;
    e = rb_exc_new2(ec, buf);
    rb_iv_set(e, "rc", INT2NUM(rc));
    rb_exc_raise(e);
}

/*
 * call-seq:
 *  create(path, option=>value, ...)
 *
 * インデックスファイルの作成。
 *
 * 有効な option:
 * :key_size [Integer] :: キーのサイズ (default: 0)
 * :flags [Integer] :: 次のフラグを「|」で連結したもの。Senna::Index::NORMALIZE, Senna::Index::SPLIT_ALPHA, Senna::Index::SPLIT_DIGIT, Senna::Index::SPLIT_SYMBOL, Senna::Index::NGRAM, Senna::Index::DELIMITED. (default: 0)
 * :initial_n_segments [Integer] :: 初期バッファサイズ (default: 0)
 * :encoding [Integer] :: 次のいずれか。Senna::ENC_DEFAULT(default), Senna::ENC_NONE, Senna::ENC_EUC_JP, Senna::ENC_UTF8, Senna::ENC_SJIS
 *
 * === Argument
 * path [String] :: インデックスファイル名
 *
 * === Return
 * Senna::Index オブジェクト
 *
 * === Exception
 * Senna::Error:: インデックスの作成に失敗した
 */
static VALUE index_create(int argc, VALUE *argv, VALUE klass)
{
    VALUE path, opt;
    VALUE key_size=Qnil, flags, initial_n_segments, encoding;
    char *s_path;
    int _key_size=0, _flags=0, _initial_n_segments=0;
    sen_encoding _encoding=sen_enc_default;
    sen_index *indexp;
    VALUE obj;
    struct index_data *data;
    rb_scan_args(argc, argv, "11", &path, &opt);
    Check_SafeStr(path);
    s_path = RSTRING_PTR(path);
    if (!NIL_P(opt)) {
        Check_Type(opt, T_HASH);
        key_size = rb_hash_aref(opt, ID2SYM(rb_intern("key_size")));
        if (key_size == rb_cInteger)
            _key_size = sizeof(long);
        else {
            _key_size = NIL_P(key_size) ? 0 : NUM2INT(key_size);
            key_size = Qnil;
        }
        flags = rb_hash_aref(opt, ID2SYM(rb_intern("flags")));
        _flags = NIL_P(flags) ? 0 : NUM2INT(flags);
        initial_n_segments = rb_hash_aref(opt, ID2SYM(rb_intern("initial_n_segments")));
        _initial_n_segments = NIL_P(initial_n_segments) ? 0 : NUM2INT(initial_n_segments);
        encoding = rb_hash_aref(opt, ID2SYM(rb_intern("encoding")));
        _encoding = NIL_P(encoding) ? sen_enc_default : NUM2INT(encoding);
    }
    indexp = sen_index_create(s_path, _key_size, _flags, _initial_n_segments, _encoding);
    if (indexp == NULL)
        senna_error(0, "cannot create: %s: %s", s_path, strerror(errno));
    obj = Data_Make_Struct(klass, struct index_data, 0, free_index, data);
    data->closed = 0;
    data->indexp = indexp;
    rb_iv_set(obj, "key_size", INT2NUM(_key_size));
    rb_iv_set(obj, "key_type", key_size);
    return obj;
}

/*
 * call-seq:
 *  new(path)
 *  open(path)
 *
 * インデックスのオープン。
 *
 * === Argument
 * path [String] :: インデックスファイル名
 *
 * === Return
 * Senna::Index オブジェクト
 *
 * === Exception
 * Senna::Error:: インデックスのオープンに失敗した
 */
static VALUE index_open(int argc, VALUE *argv, VALUE klass)
{
    VALUE path, opt;
    VALUE key_size = Qnil;
    char *s_path;
    sen_index *indexp;
    VALUE obj;
    int _key_size;
    struct index_data *data;

    rb_scan_args(argc, argv, "11", &path, &opt);
    Check_SafeStr(path);
    if (!NIL_P(opt)) {
        Check_Type(opt, T_HASH);
        key_size = rb_hash_aref(opt, ID2SYM(rb_intern("key_size")));
    }
    s_path = RSTRING_PTR(path);
    if ((indexp = sen_index_open(s_path)) == NULL)
        senna_error(0, "cannot open: %s: %s", s_path, strerror(errno));
    obj = Data_Make_Struct(klass, struct index_data, 0, free_index, data);
    data->closed = 0;
    data->indexp = indexp;
    sen_index_info(indexp, &_key_size, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    if (!NIL_P(key_size)) {
        int ks;
        if (key_size == rb_cInteger)
            ks = sizeof(long);
        else {
            ks = NIL_P(key_size) ? 0 : NUM2INT(key_size);
            key_size = Qnil;
        }
        if (ks != _key_size)
            senna_error(0, "key_size mismatch: %d != %d", ks, _key_size);
    }
    rb_iv_set(obj, "key_size", INT2NUM(_key_size));
    rb_iv_set(obj, "key_type", key_size);
    return obj;
}

/*
 * call-seq:
 *  remove(path)
 *
 * インデックスファイルの削除。
 *
 * === Argument
 * path [String] :: インデックスファイル名
 *
 * === Exception
 * Senna::Error:: インデックスの削除に失敗した
 */
static VALUE index_remove(VALUE klass, VALUE path)
{
    char *s_path;
    sen_rc rc;
    Check_SafeStr(path);
    s_path = RSTRING_PTR(path);
    if ((rc = sen_index_remove(s_path)) != sen_success)
        senna_error(rc, "cannot remove %s: %s", s_path, strerror(errno));
    return Qnil;
}

/*
 * call-seq:
 *  rename(old_path, new_path)
 *
 * インデックスファイルのリネーム。
 *
 * === Argument
 * old_path [String] :: 元ファイル名
 * new_path [Stirng] :: 新ファイル名
 *
 * === Exception
 * Senna::Error:: インデックスのリネームに失敗した
 */
static VALUE index_rename(VALUE klass, VALUE old_path, VALUE new_path)
{
    char *s_old_path;
    char *s_new_path;
    sen_rc rc;
    Check_SafeStr(old_path);
    Check_SafeStr(new_path);
    s_old_path = RSTRING_PTR(old_path);
    s_new_path = RSTRING_PTR(new_path);
    if ((rc = sen_index_rename(s_old_path, s_new_path)) != sen_success)
        senna_error(rc, "cannot rename %s to %s", s_old_path, s_new_path);
    return Qnil;
}

/*
 * インデックスのクローズ。
 *
 * === Exception
 * Senna::Error:: インデックスのクローズに失敗した
 */
static VALUE index_close(VALUE obj)
{
    struct index_data *data;
    sen_rc rc;
    Data_Get_Struct(obj, struct index_data, data);
    if (data->closed)
        rb_raise(rb_eRuntimeError, "already closed");
    if ((rc = sen_index_close(data->indexp)) != sen_success)
        senna_error(rc, "cannot close");
    data->closed = 1;
    return Qnil;
}

static VALUE values_new(VALUE klass);
static VALUE values_add(VALUE obj, VALUE str, VALUE weight);

/*
 * call-seq:
 *  update(key, old_value, new_value, option={})
 *
 * レコードを追加する場合は old_value に nil を指定する。
 * レコードを削除する場合は new_value に nil を指定する。
 * レコードを更新する場合は old_value, new_value の両方を指定する。
 *
 * 有効な option:
 * :section [Integer] :: セクション番号
 *
 * === Argument
 * key [String/Integer] :: インデックス値
 * old_value [String/Senna::Values] :: 古いレコード
 * new_value [String/Senna::Values] :: 新しいレコード
 *
 * === Exception
 * RuntimeError :: インデックスがクローズ済み
 * RuntimeError :: key のサイズがインデックスファイルのものとあっていない
 * TypeError :: open() で key_size に Integer が指定されているのに、key が整数でない
 * TypeError :: old_value, new_value が String でも Senna::Values でもない
 */
static VALUE index_update(int argc, VALUE *argv, VALUE obj)
{
    VALUE key, oldv, newv, opt, section;
    struct index_data *data;
    sen_rc rc;
    char *_key;
    long n;
    VALUE skey;
    int key_size;
    VALUE key_type;
    sen_values *_oldv=NULL, *_newv=NULL;
    unsigned int _section=1;

    rb_scan_args(argc, argv, "31", &key, &oldv, &newv, &opt);
    Data_Get_Struct(obj, struct index_data, data);
    if (data->closed)
        rb_raise(rb_eRuntimeError, "already closed");
    key_size = NUM2INT(rb_iv_get(obj, "key_size"));
    key_type = rb_iv_get(obj, "key_type");
    if (!NIL_P(key_type)) {
        if (!rb_obj_is_kind_of(key, key_type))
            rb_raise(rb_eTypeError, "wrong argument type %s (expected %s)", rb_obj_classname(key), rb_class_name(key_type));
    }
    if (key_size == 0) {
        skey = StringValue(key);
        _key = RSTRING_PTR(skey);
    } else {
        switch(TYPE(key)) {
        case T_FIXNUM:
        case T_BIGNUM:
            if (key_size != sizeof(n))
                rb_raise(rb_eRuntimeError, "invalid key_size: %d != %d", key_size, sizeof(n));
            n = NUM2LONG(key);
            _key = (char *)&n;
            break;
        default:
            skey = StringValue(key);
            if (key_size != RSTRING_LEN(skey))
                rb_raise(rb_eRuntimeError, "invalid key_size: %d != %d", key_size, RSTRING_LEN(skey));
            _key = RSTRING_PTR(skey);
            break;
        }
    }
    if (!NIL_P(oldv)) {
        if (TYPE(oldv) == T_STRING) {
            VALUE str = oldv;
            oldv = values_new(cSennaValues);
            values_add(oldv, str, INT2FIX(0));
        } else {
            if (!rb_obj_is_kind_of(oldv, cSennaValues))
                rb_raise(rb_eTypeError, "wrong argument type %s (expected Senna::Values)", rb_obj_classname(oldv));
        }
        Data_Get_Struct(oldv, sen_values, _oldv);
    }
    if (!NIL_P(newv)) {
        if (TYPE(newv) == T_STRING) {
            VALUE str = newv;
            newv = values_new(cSennaValues);
            values_add(newv, str, INT2FIX(0));
        } else {
            if (!rb_obj_is_kind_of(newv, cSennaValues))
                rb_raise(rb_eTypeError, "wrong argument type %s (expected Senna::Values)", rb_obj_classname(newv));
        }
        Data_Get_Struct(newv, sen_values, _newv);
    }
    if (!NIL_P(opt)) {
        Check_Type(opt, T_HASH);
        section = rb_hash_aref(opt, ID2SYM(rb_intern("section")));
        _section = NIL_P(section) ? 1 : NUM2UINT(section);
    }
    rc = sen_index_update(data->indexp, _key, _section, _oldv, _newv);
    if (rc != sen_success)
        senna_error(rc, "cannot update: key=%s", _key);
    return obj;
}

/*
 * call-seq:
 *  select(string, option=>value, ...)
 *  select(string, recs, option=>value, ...)
 *  select(string, recs, op, option=>value, ...)
 *
 * インデックスの検索。
 *
 * 有効な option:
 * :mode [Integer] :: 次のいずれか。Senna::SEL_EXACT, Senna::SEL_PARTIAL, Senna::SEL_UNSPLIT, Senna::SEL_NEAR, Senna::SEL_SIMILAR, Senna::SEL_PREFIX, Senna::SEL_SUFFIX
 * :similarity_threshold [Integer] :: Senna API DOC 参照
 * :max_interval [Integer] :: Senna API DOC 参照
 * :weight_vector [Array of Integer] :: Senna API DOC 参照
 * :vector_size [Integer] :: Senna API DOC 参照
 *
 * === Argument
 * string [String] :: 検索する文字列
 * recs [Senna::Records] :: 結果が反映されるレコード群。未指定時は空の Senna::Records が作成される。
 * op [Integer] :: 次のいずれか。Senna::SEL_OR(default), Senna::SEL_AND, Senna::SEL_BUT, Senna::SEL_ADJUST
 *
 * === Return
 * Senna::Records :: recs と同じ。
 */
static VALUE index_select(int argc, VALUE *argv, VALUE obj)
{
    VALUE string;
    VALUE arg1, arg2, arg3;
    VALUE recs = Qnil;
    VALUE op = Qnil;
    VALUE optarg = Qnil;
    struct index_data *data;
    sen_records *_recs;
    sen_sel_operator _op;
    sen_select_optarg _optarg, *_optarg_p;
    sen_rc rc;
    int n;
    n = rb_scan_args(argc, argv, "13", &string, &arg1, &arg2, &arg3);
    Check_Type(string, T_STRING);
    switch (n) {
    case 1:
        break;
    case 2:
        if (TYPE(arg1) == T_HASH)
            optarg = arg1;
        else
            recs = arg1;
        break;
    case 3:
        recs = arg1;
        if (TYPE(arg2) == T_HASH)
            optarg = arg2;
        else
            op = arg2;
        break;
    case 4:
        recs = arg1;
        op = arg2;
        optarg = arg3;
        break;
    }
    if (NIL_P(recs)) {
        _recs = sen_records_open(sen_rec_document, sen_rec_none, 0);
        recs = Data_Wrap_Struct(cSennaRecords, 0, sen_records_close, _recs);
    } else {
        if (!rb_obj_is_kind_of(recs, cSennaRecords))
            rb_raise(rb_eTypeError, "wrong argument type %s (expected Senna::Records)", rb_obj_classname(recs));
        Data_Get_Struct(recs, sen_records, _recs);
    }
    _op = NIL_P(op) ? sen_sel_or : NUM2INT(op);
    if (NIL_P(optarg)) {
        _optarg_p = NULL;
    } else {
        VALUE mode, st, mi, wv, vs;
        sen_sel_mode _mode;
        int _st, _mi, *_wv, _vs;
        int i;
        Check_Type(optarg, T_HASH);
        mode = rb_hash_aref(optarg, ID2SYM(rb_intern("mode")));
        _mode = NIL_P(mode) ? sen_sel_exact : NUM2INT(mode);
        st = rb_hash_aref(optarg, ID2SYM(rb_intern("similarity_threshold")));
        _st = NIL_P(st) ? 0 : NUM2INT(st);
        mi = rb_hash_aref(optarg, ID2SYM(rb_intern("max_interval")));
        _mi = NIL_P(mi) ? 0 : NUM2INT(mi);
        wv = rb_hash_aref(optarg, ID2SYM(rb_intern("weight_vector")));
        if (!NIL_P(wv)) {
            Check_Type(wv, T_ARRAY);
            _wv = xmalloc(RARRAY_LEN(wv) * sizeof(int));
            for (i = 0; i < RARRAY_LEN(wv); i += 1) {
                _wv[i] = NUM2INT(RARRAY_PTR(wv)[i]);
            }
            _vs = RARRAY_LEN(wv);
        } else {
            _wv = NULL;
            vs = rb_hash_aref(optarg, ID2SYM(rb_intern("vector_size")));
            _vs = NIL_P(vs) ? 0 : NUM2INT(vs);
        }
        _optarg.mode = _mode;
        _optarg.similarity_threshold = _st;
        _optarg.max_interval = _mi;
        _optarg.weight_vector = _wv;
        _optarg.vector_size = _vs;
        _optarg.func = NULL;
        _optarg_p = &_optarg;
    }
    Data_Get_Struct(obj, struct index_data, data);
    if (data->closed)
        rb_raise(rb_eRuntimeError, "already closed");
    rc = sen_index_select(data->indexp, RSTRING_PTR(string), RSTRING_LEN(string), _recs, _op, _optarg_p);
    if (rc != sen_success)
        senna_error(rc, "sen_index_select failed");
    return recs;
}

/*
 * インデックス情報の取得。
 *
 * === Return
 * Hash :: インデックス情報を表すハッシュ。次のキーを持つ。:key_size, :flags, :initial_n_segments, :encoding
 */
static VALUE index_info(VALUE obj)
{
    struct index_data *data;
    int key_size, flags, initial_n_segments;
    sen_encoding encoding;
    VALUE ret;
    Data_Get_Struct(obj, struct index_data, data);
    sen_index_info(data->indexp, &key_size, &flags, &initial_n_segments, &encoding, NULL, NULL, NULL, NULL, NULL, NULL);
    ret = rb_hash_new();
    rb_hash_aset(ret, ID2SYM(rb_intern("key_size")), INT2NUM(key_size));
    rb_hash_aset(ret, ID2SYM(rb_intern("flags")), INT2NUM(flags));
    rb_hash_aset(ret, ID2SYM(rb_intern("initial_n_segments")), INT2NUM(initial_n_segments));
    rb_hash_aset(ret, ID2SYM(rb_intern("encoding")), INT2NUM(encoding));
    return ret;
}

/*
 * call-seq:
 *  new(record_unit=Senna::Records::DOCUMENT, subrec_unit=Senna::Records::NONE, max_n_subrecs=0)
 *
 * 結果レコードオブジェクトの生成。
 *
 * === Argument
 * record_unit [Integer] :: レコードの単位。次のいずれか。Senna::Records::DOCUMENT(default), Senna::Records::SECTION, Senna::Records::POSITION, Senna::Records::USERDEF, Senna::Records::NONE(default)
 * subrec_unit [Integer] :: サブレコードの単位。指定できる値は record_unit と同じ。
 * max_n_subrecs [Integer] :: レコード内のサブレコードの最大数。
 *
 * === Return
 * Senna::Records
 */
static VALUE records_new(int argc, VALUE *argv, VALUE klass)
{
    VALUE record_unit;
    VALUE subrec_unit;
    VALUE max_n_subrecs;
    sen_records *recs;
    VALUE obj;
    sen_rec_unit _record_unit, _subrec_unit;
    int _max_n_subrecs;
    rb_scan_args(argc, argv, "03", &record_unit, &subrec_unit, &max_n_subrecs);
    _record_unit = NIL_P(record_unit) ? sen_rec_document : NUM2INT(record_unit);
    _subrec_unit = NIL_P(subrec_unit) ? sen_rec_none : NUM2INT(subrec_unit);
    _max_n_subrecs = NIL_P(max_n_subrecs) ? 0 : NUM2INT(max_n_subrecs);
    recs = sen_records_open(_record_unit, _subrec_unit, _max_n_subrecs);
    obj = Data_Wrap_Struct(klass, 0, sen_records_close, recs);
    rb_funcall(obj, rb_intern("initialize"), 3, record_unit, subrec_unit, max_n_subrecs);
    return obj;
}

static VALUE records_initialize(VALUE obj, VALUE record_unit, VALUE subrec_unit, VALUE max_n_subrecs)
{
    return Qnil;
}

/*
 * 次のレコードを読む。
 *
 * === Return
 * String:: レコードのキー
 * Integer:: レコードのスコア
 */
static VALUE records_next(VALUE obj)
{
    sen_records *recs;
    int keylen;
    char keybuf[256];
    char *bufp;
    int score;
    VALUE key;
    Data_Get_Struct(obj, sen_records, recs);
    keylen = sen_records_next(recs, keybuf, sizeof(keybuf), &score);
    if (keylen == 0)
        return Qnil;
    if (keylen <= sizeof(keybuf)) {
        key = rb_str_new(keybuf, keylen-1);
    } else {
        bufp = xmalloc(keylen);
        sen_records_curr_key(recs, bufp, keylen);
        key = rb_str_new(bufp, keylen-1);
        xfree(bufp);
    }
    return rb_ary_new3(2, key, INT2NUM(score));
}

/*
 * call-seq:
 *  each{|key,score| ... }
 *
 * レコードを繰り返す。
 * ブロック未指定時は Enumerator を返す。
 *
 * === Block Parameter
 * key [String] :: レコードのキー
 * score [Integer] :: レコードのスコア
 *
 * === Return
 * self
 */
static VALUE records_each(VALUE obj)
{
    VALUE val;
#ifdef RETURN_ENUMERATOR
    RETURN_ENUMERATOR(obj, 0, 0);
#endif
    while ((val = records_next(obj)) != Qnil) {
        rb_yield(val);
    }
    return obj;
}

/*
 * 最初のレコード位置に戻す
 *
 * === Exception
 * Senna::Error
 */
static VALUE records_rewind(VALUE obj)
{
    sen_records *recs;
    sen_rc rc;
    Data_Get_Struct(obj, sen_records, recs);
    if ((rc = sen_records_rewind(recs)) != sen_success)
        senna_error(rc, "cannot rewind");
    return Qnil;
}

/*
 * 現在のレコードのスコアを返す
 *
 * === Return
 * Integer
 */
static VALUE records_curr_score(VALUE obj)
{
    sen_records *recs;
    Data_Get_Struct(obj, sen_records, recs);
    return INT2NUM(sen_records_curr_score(recs));
}

/*
 * 現在のレコードのキーを返す
 *
 * === Return
 * String
 */
static VALUE records_curr_key(VALUE obj)
{
    sen_records *recs;
    int keylen;
    char keybuf[256];
    char *bufp;
    VALUE key;

    Data_Get_Struct(obj, sen_records, recs);
    keylen = sen_records_curr_key(recs, NULL, 0);
    if (keylen == 0)
        return Qnil;
    if (keylen <= sizeof(keybuf)) {
        key = rb_str_new(keybuf, keylen-1);
    } else {
        bufp = xmalloc(keylen+1);
        sen_records_curr_key(recs, bufp, keylen+1);
        key = rb_str_new(bufp, keylen-1);
        xfree(bufp);
    }
    return key;
}

/*
 * 含まれているレコードの数を返す。
 *
 * === Return
 * Integer
 */
static VALUE records_nhits(VALUE obj)
{
    sen_records *recs;
    Data_Get_Struct(obj, sen_records, recs);
    return INT2NUM(sen_records_nhits(recs));
}

/*
 * key に対応するレコードがあれば、そのレコードのスコアを返す
 *
 * === Argument
 * key [String] :: キー
 *
 * === Return
 * Integer
 */
static VALUE records_find(VALUE obj, VALUE key)
{
    sen_records *recs;
    Data_Get_Struct(obj, sen_records, recs);
    Check_Type(key, T_STRING);
    return INT2NUM(sen_records_find(recs, RSTRING_PTR(key)));
}

/*
 * Senna::Values オブジェクトの作成
 *
 * === Return
 * Senna::Values
 */
static VALUE values_new(VALUE klass)
{
    sen_values *values;
    VALUE obj;
    values = sen_values_open();
    obj = Data_Wrap_Struct(klass, 0, sen_values_close, values);
    rb_funcall(obj, rb_intern("initialize"), 0);
    return obj;
}

/*
 * call-seq:
 *  add(str, weight)
 *
 * 重みをつけて文字列を追加する。
 *
 * === Argument
 * str [String] :: 追加する文字列
 * weight [Integer] :: 重み
 *
 * === Return
 * self
 */
static VALUE values_add(VALUE obj, VALUE str, VALUE weight)
{
    sen_values *values;
    sen_rc rc;
    Data_Get_Struct(obj, sen_values, values);
    Check_Type(str, T_STRING);
    rc = sen_values_add(values, RSTRING_PTR(str), RSTRING_LEN(str), NUM2UINT(weight));
    if (rc != sen_success)
        senna_error(rc, "sen_values_add failed: str=%.32s", RSTRING_PTR(str));
    return obj;
}

/*
 * エラー番号を返す
 *
 * === Return
 * Integer
 */
static VALUE error_rc(VALUE obj)
{
    return rb_iv_get(obj, "rc");
}

/*
 * = Senna/Ruby
 *
 * Senna ライブラリを Ruby から使用するためのラッパーモジュール。
 *
 * == Usage
 *  # インデックスの作成。既存のインデックスを指定した場合は中身が空になる。
 *  idx = Senna::Index.create "/path/to/index", :encoding=>Senna::ENC_UTF8, :flags=>Senna::Index::NGRAM
 *  # レコード追加。
 *  idx.update "keyname", nil, "new recored data string"
 *  # クローズ。
 *  idx.close
 *
 *  # 既存インデックスのオープン。
 *  idx = Senna::Index.open "/path/to/index"
 *  # 語句を検索。
 *  recs = idx.select "data"
 *  recs.each do |key, score|
 *    puts key
 *  end
 */
 void Init_senna(void)
{
    sen_init();
    mSenna = rb_define_module("Senna");
    rb_define_const(mSenna, "ENC_DEFAULT", INT2NUM(sen_enc_default));
    rb_define_const(mSenna, "ENC_NONE", INT2NUM(sen_enc_none));
    rb_define_const(mSenna, "ENC_EUC_JP", INT2NUM(sen_enc_euc_jp));
    rb_define_const(mSenna, "ENC_UTF8", INT2NUM(sen_enc_utf8));
    rb_define_const(mSenna, "ENC_SJIS", INT2NUM(sen_enc_sjis));
    rb_define_const(mSenna, "SEL_OR", INT2NUM(sen_sel_or));
    rb_define_const(mSenna, "SEL_AND", INT2NUM(sen_sel_and));
    rb_define_const(mSenna, "SEL_BUT", INT2NUM(sen_sel_but));
    rb_define_const(mSenna, "SEL_ADJUST", INT2NUM(sen_sel_adjust));
    rb_define_const(mSenna, "SEL_EXACT", INT2NUM(sen_sel_exact));
    rb_define_const(mSenna, "SEL_PARTIAL", INT2NUM(sen_sel_partial));
    rb_define_const(mSenna, "SEL_UNSPLIT", INT2NUM(sen_sel_unsplit));
    rb_define_const(mSenna, "SEL_NEAR", INT2NUM(sen_sel_near));
    rb_define_const(mSenna, "SEL_SIMILAR", INT2NUM(sen_sel_similar));
    rb_define_const(mSenna, "SEL_PREFIX", INT2NUM(sen_sel_prefix));
    rb_define_const(mSenna, "SEL_SUFFIX", INT2NUM(sen_sel_suffix));

    /*
     * Senna インデックス
     */
    cSennaIndex = rb_define_class_under(mSenna, "Index", rb_cObject);
    rb_define_const(cSennaIndex, "NORMALIZE", INT2NUM(SEN_INDEX_NORMALIZE));
    rb_define_const(cSennaIndex, "SPLIT_ALPHA", INT2NUM(SEN_INDEX_SPLIT_ALPHA));
    rb_define_const(cSennaIndex, "SPLIT_DIGIT", INT2NUM(SEN_INDEX_SPLIT_DIGIT));
    rb_define_const(cSennaIndex, "SPLIT_SYMBOL", INT2NUM(SEN_INDEX_SPLIT_SYMBOL));
    rb_define_const(cSennaIndex, "NGRAM", INT2NUM(SEN_INDEX_NGRAM));
    rb_define_const(cSennaIndex, "DELIMITED", INT2NUM(SEN_INDEX_DELIMITED));
    rb_define_singleton_method(cSennaIndex, "create", index_create, -1);
    rb_define_singleton_method(cSennaIndex, "open", index_open, -1);
    rb_define_singleton_method(cSennaIndex, "new", index_open, -1);
    rb_define_singleton_method(cSennaIndex, "remove", index_remove, 1);
    rb_define_singleton_method(cSennaIndex, "rename", index_rename, 2);
    rb_define_method(cSennaIndex, "close", index_close, 0);
    rb_define_method(cSennaIndex, "select", index_select, -1);
    rb_define_method(cSennaIndex, "update", index_update, -1);
    rb_define_method(cSennaIndex, "info", index_info, 0);

    /*
     * 検索結果レコード群
     */
    cSennaRecords = rb_define_class_under(mSenna, "Records", rb_cObject);
    rb_define_const(cSennaRecords, "DOCUMENT", INT2NUM(sen_rec_document));
    rb_define_const(cSennaRecords, "SECTION", INT2NUM(sen_rec_section));
    rb_define_const(cSennaRecords, "POSITION", INT2NUM(sen_rec_position));
    rb_define_const(cSennaRecords, "USERDEF", INT2NUM(sen_rec_userdef));
    rb_define_const(cSennaRecords, "NONE", INT2NUM(sen_rec_none));
    rb_define_singleton_method(cSennaRecords, "new", records_new, -1);
    rb_define_method(cSennaRecords, "initialize", records_initialize, 3);
    rb_define_method(cSennaRecords, "next", records_next, 0);
    rb_define_method(cSennaRecords, "each", records_each, 0);
    rb_define_method(cSennaRecords, "rewind", records_rewind, 0);
    rb_define_method(cSennaRecords, "curr_score", records_curr_score, 0);
    rb_define_method(cSennaRecords, "curr_key", records_curr_key, 0);
    rb_define_method(cSennaRecords, "nhits", records_nhits, 0);
    rb_define_method(cSennaRecords, "find", records_find, 1);
//    rb_define_method(cSennaRecords, "close", records_close, 0);

    /*
     * 重みつき文字列を格納するためのクラス
     */
    cSennaValues = rb_define_class_under(mSenna, "Values", rb_cObject);
    rb_define_singleton_method(cSennaValues, "new", values_new, 0);
    rb_define_method(cSennaValues, "add", values_add, 2);

    /*
     * 他のエラーの基底クラス。
     * Senna ライブラリがエラーを返さないエラーの場合には Senna::Error 自体が生成される。
     */
    eSennaError = rb_define_class_under(mSenna, "Error", rb_eStandardError);
    rb_define_method(eSennaError, "rc", error_rc, 0);
    /* sen_memory_exhausted に対応するエラー */
    eSennaMemoryExhausted = rb_define_class_under(mSenna, "MemoryExhausted", eSennaError);
    /* sen_invalid_format に対応するエラー */
    eSennaInvalidFormat = rb_define_class_under(mSenna, "InvalidFormat", eSennaError);
    /* sen_file_operation_error に対応するエラー */
    eSennaFileOperationError = rb_define_class_under(mSenna, "FileOperationError", eSennaError);
    /* sen_invalid_argument に対応するエラー */
    eSennaInvalidArgument = rb_define_class_under(mSenna, "InvalidArgument", eSennaError);
    /* sen_other_error に対応するエラー */
    eSennaOtherError = rb_define_class_under(mSenna, "OtherError", eSennaError);
    /* sen_external_error に対応するエラー */
    eExternalError = rb_define_class_under(mSenna, "ExternalError", eSennaError);
    /* sen_internal_error に対応するエラー */
    eInternalError = rb_define_class_under(mSenna, "InternalError", eSennaError);
    /* sen_abnormal_error に対応するエラー */
    eAbnormalError = rb_define_class_under(mSenna, "AbnormalError", eSennaError);
    /* sen_end_of_data に対応するエラー */
    eEndOfData = rb_define_class_under(mSenna, "EndOfData", eSennaError);
}
